@@ -16,7 +16,7 @@
  * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL SuSE
  * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
- * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
+ * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
  * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  *
  * Author:  Keith Packard, SuSE, Inc.
@@ -26,6 +26,7 @@
 #include <config.h>
 #endif
 #include "Xrenderint.h"
+#include <limits.h>
 
 XRenderExtInfo XRenderExtensionInfo;
 char XRenderExtensionName[] = RENDER_NAME;
@@ -38,7 +39,7 @@ static int XRenderCloseDisplay (Display *dpy, XExtCodes *codes);
  * XextFindDisplay.)
  */
 static XRenderExtDisplayInfo *
-XRenderExtFindDisplay (XRenderExtInfo *extinfo, 
+XRenderExtFindDisplay (XRenderExtInfo *extinfo,
                        Display        *dpy)
 {
     XRenderExtDisplayInfo *dpyinfo;
@@ -46,7 +47,7 @@ XRenderExtFindDisplay (XRenderExtInfo *extinfo,
     /*
      * see if this was the most recently accessed display
      */
-    if ((dpyinfo = extinfo->cur) && dpyinfo->display == dpy) 
+    if ((dpyinfo = extinfo->cur) && dpyinfo->display == dpy)
         return dpyinfo;
 
     /*
@@ -70,8 +71,8 @@ XRenderExtFindDisplay (XRenderExtInfo *extinfo,
  * any screen, tell the application that Render is not present.
  */
 
-#define DEPTH_MASK(d)	(1 << ((d) - 1))
-    
+#define DEPTH_MASK(d)	(1U << ((d) - 1))
+
 /*
  * Render requires support for depth 1, 4, 8, 24 and 32 pixmaps
  */
@@ -81,7 +82,7 @@ XRenderExtFindDisplay (XRenderExtInfo *extinfo,
 			 DEPTH_MASK(8) | \
 			 DEPTH_MASK(24) | \
 			 DEPTH_MASK(32))
-    
+
 typedef struct _DepthCheckRec {
     struct _DepthCheckRec *next;
     Display *dpy;
@@ -204,11 +205,11 @@ XRenderExtAddDisplay (XRenderExtInfo *extinfo,
 	dpyinfo->codes = NULL;
 
     /*
-     * if the server has the extension, then we can initialize the 
+     * if the server has the extension, then we can initialize the
      * appropriate function vectors
      */
     if (dpyinfo->codes) {
-        XESetCloseDisplay (dpy, dpyinfo->codes->extension, 
+        XESetCloseDisplay (dpy, dpyinfo->codes->extension,
                            XRenderCloseDisplay);
     } else {
 	/* The server doesn't have this extension.
@@ -241,7 +242,7 @@ XRenderExtAddDisplay (XRenderExtInfo *extinfo,
  * XRenderExtRemoveDisplay - remove the indicated display from the
  * extension object. (Replaces XextRemoveDisplay.)
  */
-static int 
+static int
 XRenderExtRemoveDisplay (XRenderExtInfo *extinfo, Display *dpy)
 {
     XRenderExtDisplayInfo *dpyinfo, *prev;
@@ -285,7 +286,7 @@ XRenderFindDisplay (Display *dpy)
 
     dpyinfo = XRenderExtFindDisplay (&XRenderExtensionInfo, dpy);
     if (!dpyinfo)
-	dpyinfo = XRenderExtAddDisplay (&XRenderExtensionInfo, dpy, 
+	dpyinfo = XRenderExtAddDisplay (&XRenderExtensionInfo, dpy,
                                         XRenderExtensionName);
     return dpyinfo;
 }
@@ -295,10 +296,10 @@ XRenderCloseDisplay (Display *dpy, XExtCodes *codes)
 {
     XRenderExtDisplayInfo *info = XRenderFindDisplay (dpy);
     if (info && info->info) XFree (info->info);
-    
+
     return XRenderExtRemoveDisplay (&XRenderExtensionInfo, dpy);
 }
-    
+
 /****************************************************************************
  *                                                                          *
  *			    Render public interfaces                        *
@@ -331,8 +332,8 @@ Status XRenderQueryVersion (Display *dpy,
 
     if (!XRenderQueryFormats (dpy))
 	return 0;
-    
-    xri = info->info; 
+
+    xri = info->info;
     *major_versionp = xri->major_version;
     *minor_versionp = xri->minor_version;
     return 1;
@@ -342,7 +343,7 @@ static XRenderPictFormat *
 _XRenderFindFormat (XRenderInfo *xri, PictFormat format)
 {
     int	nf;
-    
+
     for (nf = 0; nf < xri->nformat; nf++)
 	if (xri->format[nf].id == format)
 	    return &xri->format[nf];
@@ -360,7 +361,7 @@ typedef struct _renderVersionState {
     Bool	    error;
     int		    major_version;
     int		    minor_version;
-    
+
 } _XrenderVersionState;
 
 static Bool
@@ -411,9 +412,9 @@ XRenderQueryFormats (Display *dpy)
     CARD32			*xSubpixel;
     void			*xData;
     int				nf, ns, nd, nv;
-    int				rlength;
-    int				nbytes;
-    
+    unsigned long		rlength;
+    unsigned long		nbytes;
+
     RenderCheckExtension (dpy, info, 0);
     LockDisplay (dpy);
     if (info->info)
@@ -426,19 +427,19 @@ XRenderQueryFormats (Display *dpy)
     vreq->renderReqType = X_RenderQueryVersion;
     vreq->majorVersion = RENDER_MAJOR;
     vreq->minorVersion = RENDER_MINOR;
-    
+
     async_state.version_seq = dpy->request;
     async_state.error = False;
     async.next = dpy->async_handlers;
     async.handler = _XRenderVersionHandler;
     async.data = (XPointer) &async_state;
     dpy->async_handlers = &async;
-    
+
     GetReq (RenderQueryPictFormats, req);
     req->reqType = info->codes->major_opcode;
     req->renderReqType = X_RenderQueryPictFormats;
-    
-    if (!_XReply (dpy, (xReply *) &rep, 0, xFalse)) 
+
+    if (!_XReply (dpy, (xReply *) &rep, 0, xFalse))
     {
 	DeqAsyncHandler (dpy, &async);
 	UnlockDisplay (dpy);
@@ -457,25 +458,36 @@ XRenderQueryFormats (Display *dpy)
      */
     if (async_state.major_version == 0 && async_state.minor_version < 6)
 	rep.numSubpixel = 0;
-	
-    xri = (XRenderInfo *) Xmalloc (sizeof (XRenderInfo) +
-				   rep.numFormats * sizeof (XRenderPictFormat) +
-				   rep.numScreens * sizeof (XRenderScreen) +
-				   rep.numDepths * sizeof (XRenderDepth) +
-				   rep.numVisuals * sizeof (XRenderVisual));
-    rlength = (rep.numFormats * sizeof (xPictFormInfo) +
-	       rep.numScreens * sizeof (xPictScreen) +
-	       rep.numDepths * sizeof (xPictDepth) +
-	       rep.numVisuals * sizeof (xPictVisual) +
-	       rep.numSubpixel * 4);
-    xData = (void *) Xmalloc (rlength);
-    nbytes = (int) rep.length << 2;
-    
+
+    if ((rep.numFormats < ((INT_MAX / 4) / sizeof (XRenderPictFormat))) &&
+	(rep.numScreens < ((INT_MAX / 4) / sizeof (XRenderScreen))) &&
+	(rep.numDepths  < ((INT_MAX / 4) / sizeof (XRenderDepth))) &&
+	(rep.numVisuals < ((INT_MAX / 4) / sizeof (XRenderVisual))) &&
+	(rep.numSubpixel < ((INT_MAX / 4) / 4)) &&
+	(rep.length < (INT_MAX >> 2)) ) {
+	xri = Xmalloc (sizeof (XRenderInfo) +
+		       (rep.numFormats * sizeof (XRenderPictFormat)) +
+		       (rep.numScreens * sizeof (XRenderScreen)) +
+		       (rep.numDepths * sizeof (XRenderDepth)) +
+		       (rep.numVisuals * sizeof (XRenderVisual)));
+	rlength = ((rep.numFormats * sizeof (xPictFormInfo)) +
+		   (rep.numScreens * sizeof (xPictScreen)) +
+		   (rep.numDepths * sizeof (xPictDepth)) +
+		   (rep.numVisuals * sizeof (xPictVisual)) +
+		   (rep.numSubpixel * 4));
+	xData = Xmalloc (rlength);
+	nbytes = (unsigned long) rep.length << 2;
+    } else {
+	xri = NULL;
+	xData = NULL;
+	rlength = nbytes = 0;
+    }
+
     if (!xri || !xData || nbytes < rlength)
     {
 	if (xri) Xfree (xri);
 	if (xData) Xfree (xData);
-	_XEatData (dpy, nbytes);
+	_XEatDataWords (dpy, rep.length);
 	UnlockDisplay (dpy);
 	SyncHandle ();
 	return 0;
@@ -538,7 +550,7 @@ XRenderQueryFormats (Display *dpy)
 	    xDepth = (xPictDepth *) xVisual;
 	}
 	screen++;
-	xScreen = (xPictScreen *) xDepth;	    
+	xScreen = (xPictScreen *) xDepth;
     }
     xSubpixel = (CARD32 *) xScreen;
     screen = xri->screen;
@@ -554,7 +566,7 @@ XRenderQueryFormats (Display *dpy)
      */
     if (nbytes > rlength)
 	_XEatData (dpy, (unsigned long) (nbytes - rlength));
-    
+
     UnlockDisplay (dpy);
     SyncHandle ();
     Xfree (xData);
@@ -696,7 +708,7 @@ XRenderFindStandardFormat (Display  *dpy,
 		},
 		0,			    /* colormap */
 	    },
-	    PictFormatType | 
+	    PictFormatType |
 	    PictFormatDepth |
 	    PictFormatRed |
 	    PictFormatRedMask |
@@ -725,7 +737,7 @@ XRenderFindStandardFormat (Display  *dpy,
 		},
 		0,			    /* colormap */
 	    },
-	    PictFormatType | 
+	    PictFormatType |
 	    PictFormatDepth |
 	    PictFormatRed |
 	    PictFormatRedMask |
@@ -753,7 +765,7 @@ XRenderFindStandardFormat (Display  *dpy,
 		},
 		0,			    /* colormap */
 	    },
-	    PictFormatType | 
+	    PictFormatType |
 	    PictFormatDepth |
 	    PictFormatRedMask |
 	    PictFormatGreenMask |
@@ -779,7 +791,7 @@ XRenderFindStandardFormat (Display  *dpy,
 		},
 		0,			    /* colormap */
 	    },
-	    PictFormatType | 
+	    PictFormatType |
 	    PictFormatDepth |
 	    PictFormatRedMask |
 	    PictFormatGreenMask |
@@ -805,7 +817,7 @@ XRenderFindStandardFormat (Display  *dpy,
 		},
 		0,			    /* colormap */
 	    },
-	    PictFormatType | 
+	    PictFormatType |
 	    PictFormatDepth |
 	    PictFormatRedMask |
 	    PictFormatGreenMask |
@@ -816,7 +828,7 @@ XRenderFindStandardFormat (Display  *dpy,
     };
 
     if (0 <= format && format < PictStandardNUM)
-	return XRenderFindFormat (dpy, 
+	return XRenderFindFormat (dpy,
 				  standardFormats[format].mask,
 				  &standardFormats[format].templ,
 				  0);
@@ -832,7 +844,7 @@ XRenderQueryPictIndexValues(Display			*dpy,
     xRenderQueryPictIndexValuesReq	*req;
     xRenderQueryPictIndexValuesReply	rep;
     XIndexValue				*values;
-    int					nbytes, nread, rlength, i;
+    unsigned int			nbytes, nread, rlength, i;
 
     RenderCheckExtension (dpy, info, NULL);
 
@@ -848,18 +860,25 @@ XRenderQueryPictIndexValues(Display			*dpy,
 	return NULL;
     }
 
-    /* request data length */
-    nbytes = (long)rep.length << 2;
-    /* bytes of actual data in the request */
-    nread = rep.numIndexValues * SIZEOF (xIndexValue);
-    /* size of array returned to application */
-    rlength = rep.numIndexValues * sizeof (XIndexValue);
+    if ((rep.length < (INT_MAX >> 2)) &&
+	(rep.numIndexValues < (INT_MAX / sizeof (XIndexValue)))) {
+	/* request data length */
+	nbytes = rep.length << 2;
+	/* bytes of actual data in the request */
+	nread = rep.numIndexValues * SIZEOF (xIndexValue);
+	/* size of array returned to application */
+	rlength = rep.numIndexValues * sizeof (XIndexValue);
+
+	/* allocate returned data */
+	values = Xmalloc (rlength);
+    } else {
+	nbytes = nread = rlength = 0;
+	values = NULL;
+    }
 
-    /* allocate returned data */
-    values = (XIndexValue *)Xmalloc (rlength);
     if (!values)
     {
-	_XEatData (dpy, nbytes);
+	_XEatDataWords (dpy, rep.length);
 	UnlockDisplay (dpy);
 	SyncHandle ();
 	return NULL;
@@ -870,7 +889,7 @@ XRenderQueryPictIndexValues(Display			*dpy,
     for(i = 0; i < rep.numIndexValues; i++)
     {
 	xIndexValue value;
-	
+
 	_XRead (dpy, (char *) &value, SIZEOF (xIndexValue));
 	values[i].pixel = value.pixel;
 	values[i].red = value.red;
